# The following are the default redirections from Offpunk
# Those are by default because they should make sens with offpunk

redirects = {
    "*reddit.com"  : "teddit.net",
    "*medium.com"  : "scribe.rip",
    }


#following are blocked URLs. Visiting them with offpunk doesn’t make sense.
#Blocking them will save a lot of bandwith

blocked = {
    "twitter.com",
    "x.com",
#    "youtube.com",
#    "youtu.be",
    "facebook.com",
    "facebook.net",
    "fbcdn.net",
    "linkedin.com",
    "*licdn.com",
    "*admanager.google.com",
    "*google-health-ads.blogspot.com",
    "*firebase.google.com",
    "*google-webfonts-helper.herokuapp.com",
    "*tiktok.com"   ,
    "*doubleclick.net",
    "*google-analytics.com" ,
    "*ads.yahoo.com",
    "*advertising.amazon.com",
    "*advertising.theguardian.com",
    "*advertise.newrepublic.com",
}
